(* uses src/missing,
        src/psi,
        src/simulator,
        src/bisim,
        src/command
*)

signature PSI_INSTANCE_NOM =
sig
  eqtype name
  type term
  type condition
  type assertion

  val chaneq   : term * term -> condition
  val compose  : assertion * assertion -> assertion
  val unit     : assertion
  val entails  : assertion * condition -> bool

  val substT   : (name * term) list -> term      -> term
  val substC   : (name * term) list -> condition -> condition
  val substA   : (name * term) list -> assertion -> assertion

  val new      : name list -> name

  val eqT      : (((name * term) * (name * term)) -> bool) -> term * term -> bool
  val eqC      : (((name * condition) * (name * condition)) -> bool) -> condition * condition -> bool
  val eqA      : (((name * assertion) * (name * assertion)) -> bool) -> assertion * assertion -> bool

  val supportT : term -> name list
  val supportC : condition -> name list
  val supportA : assertion -> name list

  val swapT    : name * name -> term -> term
  val swapC    : name * name -> condition -> condition
  val swapA    : name * name -> assertion -> assertion
end;

signature SYMBOLIC_PSI_FLAT =
sig
  include PSI_INSTANCE_NOM

  val var : name -> term


  structure Constraint : SYMBOLIC_TRANSITION_CONSTRAINT
  val solve :
    Constraint.constraint ->
           ( condition list list
           , ((name * term) list * assertion) list (* solutions *)
           ) Either.either


  structure BisimConstraint : SYMBOLIC_BISIM_CONSTRAINT
  val solveBisim :
    ( (string -> BisimConstraint.constraint -> unit) ->
      BisimConstraint.constraint ->
          ( condition list list ,
          ((name * term) list * assertion) list)
      Either.either ) option


  sharing type Constraint.Inst.name      = name
  sharing type Constraint.Inst.term      = term
  sharing type Constraint.Inst.condition = condition
  sharing type Constraint.Inst.assertion = assertion
  sharing Constraint.Inst = BisimConstraint.Psi.Inst
end;


signature C_PSI =
sig
  include SYMBOLIC_PSI_FLAT

  val parseN : string -> (string, name) Either.either
  val parseT : string -> (string, term) Either.either
  val parseC : string -> (string, condition) Either.either
  val parseA : string -> (string, assertion) Either.either

  val printN  : name      -> string
  val printT  : term      -> string
  val printC  : condition -> string
  val printA  : assertion -> string

end;




functor PsiInstance(I: PSI_INSTANCE_NOM) : PSI_INSTANCE =
struct
  open I

  structure Term = Nominal(
  struct
    type atom   = name
    type data   = term
    val swap    = swapT
    val support = supportT
    val new     = new
    val eq      = eqT
  end)

  structure Cond = Nominal(
  struct
    type atom   = name
    type data   = condition
    val swap    = swapC
    val support = supportC
    val new     = new
    val eq      = eqC
  end)

  structure Assr = Nominal(
  struct
    type atom   = name
    type data   = assertion
    val swap    = swapA
    val support = supportA
    val new     = new
    val eq      = eqA
  end)
end;

functor Constraint(S: PSI_INSTANCE_NOM): SYMBOLIC_TRANSITION_CONSTRAINT =
struct
  structure PsiInst    = PsiInstance(S)
  structure Constraint = SymbolicOSConstraint(PsiInst)
  open Constraint
end;

functor BisimConstraint(S: PSI_INSTANCE_NOM): SYMBOLIC_BISIM_CONSTRAINT =
struct
  structure PsiInst    = PsiInstance(S)
  structure Psi        = Psi(PsiInst)
  structure Constraint = SymBisimConstraint(Psi)
  open Constraint
end;


functor Command(C: C_PSI) =
struct
  structure CP = CommandParser(struct

    (*
    structure SI = SymbolicPsi(C)
    *)
    structure Psi = C.BisimConstraint.Psi
    structure Inst = Psi.Inst
    structure SI = struct
      (*
      structure PsiInstance = C.Constraint.Inst
      structure Constraint  = C.Constraint
      *)
      structure PsiInstance = Inst
      structure Constraint  = C.Constraint
      structure Psi         = Psi
      structure Clause      = PsiClause(Psi)
      structure ClEnv       = PsiClauseEnvironment(Clause)
      val var = C.var
    end;

    structure SCS = struct
      structure Inst       = Inst
      structure Constraint = C.Constraint
      val solve            = C.solve
    end
    structure SBCS = struct
      structure Psi        = Psi
      structure Constraint = C.BisimConstraint
      val solve =
        case C.solveBisim of
             NONE => (fn pp => fn c => Either.LEFT [])
           | SOME s => s
    end
    val useBisim = case C.solveBisim of NONE => false | _ => true

    structure PsiParserReq = struct
      structure Psi = Psi
      val parseName = C.parseN
      val parseTerm = C.parseT
      val parseCond = C.parseC
      val parseAssr = C.parseA
      val var       = SOME C.var
    end
    structure PPInst = struct
      structure Inst = Psi.Inst
      val printN    = C.printN
      val printT    = C.printT
      val printC    = C.printC
      val printA    = C.printA
      val var       = SOME C.var
    end
  end)

  val start = CP.start
end;

