(* uses src/missing *)

signature NOMINAL =
sig
  (*
   * Sorts S ::= A      sorts of atoms
   *          |  D      sorts of data
   *          |  [A]S   sorts of atom abstraction
   *)

  eqtype atom (* names *)
  type data (* terms *)

  val swap_name : atom * atom -> atom -> atom
  val swap      : atom * atom -> data -> data
  val support   : data -> atom list
  val fresh     : atom -> data -> bool
  val new       : atom list -> atom             (* ax F4 *)

  val eqD       : data * data -> bool
end;


signature NOMINAL_MINIMAL =
sig
  eqtype atom
  type   data

  val swap      : atom * atom -> data -> data
  val support   : data -> atom list
  val new       : atom list -> atom

  val eq  : (((atom * data) * (atom * data)) -> bool) ->
            data * data -> bool
end;


functor Nominal(N : NOMINAL_MINIMAL): NOMINAL =
struct
  type atom = N.atom
  type data = N.data

  fun swap_name (a, b) d =
    if a = d then b else
    if b = d then a else d

  val swap    = N.swap
  val support = N.support
  val new     = N.new

  fun fresh a d = not (Lst.member a (support d))

  (* A1  ( Alpha equivalence )
    a.x = a'.x'  <=>  (a = a' /\ x = x')  \/  (a' # x /\ x' = (a a')ox)
  *)
  fun alphaEq ((a,x), (a',x')) =
    if a = a'
        then eqD (x,x')
        else let
                val a'' = new ([a,a'] @ (support x) @ (support x'))
             in
               eqD (swap (a,a'') x, swap (a',a'')  x')
             end

    (* (a = a' andalso eqD (x,x'))
    orelse
    ((fresh a' x) andalso eqD (x', swap (a, a') x))
    *)
  and eqD (d1,d2) = N.eq alphaEq (d1,d2)

end;



functor NominalName(N: sig
  eqtype atom
  val new : atom list -> atom end): NOMINAL =
struct
  type atom = N.atom
  type data = N.atom

  fun swap_name (a, b) d =
    if a = d then b else
    if b = d then a else d

  fun support a = [a]
  fun fresh a b = a <> b
  val swap      = swap_name
  val new       = N.new

  fun eqD (n,m) = n = m
end;


functor NominalList (A: NOMINAL): NOMINAL =
struct
  type atom     = A.atom
  type data     = A.data list

  val swap_name = A.swap_name
  fun swap ab l = map (A.swap ab) l
  fun support l = List.concat (map A.support l)
  fun fresh a l = List.all (A.fresh a) l
  val new       = A.new

  fun eqD ([], []) = true
    | eqD (n::ns, m::ms) =
        List.length ns = List.length ms andalso
        A.eqD (n,m) andalso eqD (ns, ms)
    | eqD _ = false
end;


functor NominalNameList (N: sig
  eqtype atom
  val new : atom list -> atom end) =
    NominalList (NominalName (N));


functor NominalAlpha(N: NOMINAL):
sig
  val makeFresh  : N.atom list -> N.data -> N.atom list -> N.data
  val permute    : (N.atom * N.atom) list -> N.data -> N.data
  val freshNames : N.atom list -> N.atom list -> (N.atom * N.atom) list
end =
struct
  structure L = NominalNameList (struct
    type atom = N.atom val
    new = N.new end)

  fun permute pi t = Lst.foldl N.swap t pi

  fun freshNames names support =
  let
    (* right most binders (sequence) binds the tightest,
     * i.e. right associative, as genPI generates permutations
     * in a reverse order foldl is appropriate here *)
    val dirty = List.filter (not o (Fun.flip L.fresh support)) names

    fun genPI pi [] _ = pi
      | genPI pi (d::dirty) supp =
      let val n = N.new supp
      in genPI ((d,n)::pi) dirty (n::supp)
      end

    val pi = genPI [] dirty support
  in
    pi
  end

  fun makeFresh binders term forNames =
    permute (freshNames binders forNames) term

end;

